# -*- coding: utf-8 -*-

from buildbot.schedulers.basic import (SingleBranchScheduler,
                                       AnyBranchScheduler)
from toxicbuild.schedulers.base import addBuildsetForChanges


class ToxicSingleBranchScheduler(SingleBranchScheduler):
    addBuildsetForChanges = addBuildsetForChanges


class ToxicAnyBranchScheduler(AnyBranchScheduler):
    addBuildsetForChanges = addBuildsetForChanges



# -*- coding: utf-8 -*-

import os
from buildbot import interfaces
from buildbot.steps.shell import ShellCommand
from buildbot.steps.source.git import Git
from toxicbuild.exceptions import CandyNotFound


class Candy(object):
    """
    Candies are collections of steps for common tasks.
    """

    name = 'Candy'

    def __init__(self, *args, **kwargs):
        self.build = kwargs.get('build')

    def getSteps(self):
        """
        Returns a list of steps to be used a build
        """
        raise NotImplementedError

    def getEnv(self):
        """
        Returns a dictionary with the environment variables
        to be used. Use the buildbot's steps env dict syntax.
        """

        raise NotImplementedError

    @classmethod
    def getCandies(cls):
        return cls.__subclasses__()

    @classmethod
    def getCandy(cls, candy_name):
        for candy in cls.getCandies():
            if candy.name == candy_name:
                return candy

        raise CandyNotFound('Candy %s does not exist' % candy_name)


class PythonVirtualenv(Candy):
    """
    Candy for python projects using virtualenv.
    It has steps to create a new virtualenv and install
    dependencies using pip and a requirements.txt file.
    """

    name = 'python-virtualenv'

    def __init__(self, **kwargs):
        if 'venv_path' not in kwargs or 'pyversion' not in kwargs:
            raise Exception('venv_path and pyversion params are mandatory')

        self.venv_path = kwargs.get('venv_path')
        self.pyversion = kwargs.get('pyversion')

    def getSteps(self):
        env = self.getEnv()
        create_env = ShellCommand(
            name=str('Creating virtualenv at %s' % self.venv_path),
            command=['virtualenv', '%s' % self.venv_path,
                     '--python=%s' % self.pyversion],
            env=env)

        install_deps = ShellCommand(
            name=str('Installing dependencies with pip'),
            command=['pip', 'install', '-r', 'requirements.txt'],
            env=env)

        steps = [interfaces.IBuildStepFactory(create_env),
                 interfaces.IBuildStepFactory(install_deps)]

        return steps

    def getEnv(self):
        bin_dir = os.path.join(self.venv_path, 'bin')
        env = {'PATH': [bin_dir, '${PATH}']}
        return env


class GitUpdateAndCheckout(Candy):
    name = 'git-update-and-checkout'

    def getSteps(self):
        update = Git(
            name=str("Updating code at %s" % self.build.repourl),
            repourl=self.build.repourl,
            mode='incremental')

        checkout = ShellCommand(
            name=str("Checking out to %s" % self.build.named_tree),
            command=['git', 'checkout', '%s' % self.build.named_tree])

        steps = [interfaces.IBuildStepFactory(update),
                 interfaces.IBuildStepFactory(checkout)]

        return steps



# -*- coding: utf-8 -*-

from twisted.python import log
from twisted.internet import defer
from twisted.application import service
from buildbot import config
from buildbot.db import connector, enginestrategy, pool
from toxicbuild.db import model
from toxicbuild.db.revisionconfig import RevisionConfigConnectorComponent


class DBConnector(connector.DBConnector):
    # lazy copy from bb connector

    def __init__(self, master, basedir):
        service.MultiService.__init__(self)
        self.setName('toxicdb')
        self.master = master
        self.basedir = basedir

        # not configured yet - we don't build an engine until the first
        # reconfig
        self.configured_url = None

        # set up components
        self._engine = None  # set up in reconfigService
        self.pool = None  # set up in reconfigService
        self.model = model.Model(self)
        self.revisionconfig = RevisionConfigConnectorComponent(self)

    def setup(self, verbose=True):
        db_url = self.configured_url = self.master.config.db[
            'toxicbuild_db_url']

        log.msg("Setting up database with URL %r" % (db_url,))

        # set up the engine and pool
        self._engine = enginestrategy.create_engine(db_url,
                                                    basedir=self.basedir)
        self.pool = pool.DBThreadPool(self._engine, verbose=verbose)

        # oh, yeah! all that nice check_version stuff is missing...
        return defer.succeed(None)

    def reconfigService(self, new_config):
        # double-check -- the master ensures this in config checks
        assert self.configured_url == new_config.db['toxicbuild_db_url']

        return config.ReconfigurableServiceMixin.reconfigService(self,
                                                                 new_config)



# -*- coding: utf-8 -*-

from buildbot import interfaces
from buildbot.process.build import Build
from buildbot.steps.shell import ShellCommand
from toxicbuild import master
from toxicbuild.config import ConfigReader
from toxicbuild.candies import Candy


class DynamicBuild(Build):
    """
    Build that creates per-buid config based steps.
    """
    def __init__(self, *args, **kwargs):
        self.builder_dict = {}
        self.repourl = None
        # named_tree is a branch or revision
        self.named_tree = None
        Build.__init__(self, *args, **kwargs)

    def setupBuild(self, *args, **kwargs):
        revision = self.getProperty('revision')
        branch = self.getProperty('branch') or 'master'
        # I really need to block things here, otherwise
        # I would have to change lots of things to work
        # async.
        conn = master.TOXICDB.pool.engine.connect()
        revconf = master.TOXICDB.revisionconfig._getRevisionConfig(
            conn, branch, revision=revision)

        self.config = ConfigReader(revconf.config)
        self.repourl = revconf.repourl
        self.named_tree = revision or branch
        self.steps = self.getSteps()

        self.setStepFactories(self.steps)
        Build.setupBuild(self, *args, **kwargs)

    def getSteps(self):
        # I think the 'right' place for it should be in the
        # BuildFactory class but the revision is not known there
        # so I need to do it here.
        builder_config_error = False
        try:
            self.builder_dict = self.config.getBuilder(self.builder.name)
        except Exception as e:
            name = "Config Error"
            steps = [self._createBombStep(name=name, exception=e)]
            builder_config_error = True

        if builder_config_error:
            return steps

        steps = self.getCandiesSteps()

        configured_steps = self.builder_dict['steps']
        for cmd in configured_steps:
            step = self.create_step(cmd)
            steps.append(step)
        return steps

    def getCandiesSteps(self):
        steps = []
        try:
            candies = self.getCandies()
        except Exception as e:
            step_name = 'Candy config error!'
            step = BombStep(name=step_name, exception=e)
            steps.append(interfaces.IBuildStepFactory(step))
            candies = []

        for candy in candies:
            try:
                steps += candy.getSteps()
            except NotImplementedError:  # pragma: no cover
                pass

        return steps

    def getCandies(self):
        candies = []
        candies_config = self.builder_dict.get('candies') or []
        if not candies_config:
            return candies

        for cdict in candies_config:
            cname = cdict['name']
            candy_class = Candy.getCandy(cname)
            candy = candy_class(**cdict)
            candy.build = self
            candies.append(candy)

        return candies

    def create_step(self, cmd):
        env = self.get_step_env()
        scmd = ShellCommand(env=env, **cmd)
        step = interfaces.IBuildStepFactory(scmd)
        return step

    def get_step_env(self):
        env = {}
        for candy in self.getCandies():
            try:
                env.update(candy.getEnv())
            except NotImplementedError:  # pragma: no cover
                pass

        return env

    def _createBombStep(self, name, exception):
        bomb_step = BombStep(name=name, exception=exception)
        return interfaces.IBuildStepFactory(bomb_step)


class BombStep(ShellCommand):
    """
    Make a build be marked as exception
    """
    def __init__(self, *args, **kwargs):
        try:
            self.exception = kwargs['exception']
            del kwargs['exception']
        except KeyError:
            self.exception = Exception

        ShellCommand.__init__(self, *args, **kwargs)

    def startStep(self, *args, **kwargs):
        raise self.exception



# -*- coding: utf-8 -*-

import os
import sys
from copy import copy
from buildbot.config import BuilderConfig
from buildbot.config import MasterConfig as MasterConfigBase
from toxicbuild.exceptions import BuilderNotFound


class ConfigReader(object):
    def __init__(self, conf_in):
        try:
            self.config = self._read_config(conf_in)
            self.builders = self.getBuilders()
        except Exception as e:
            self.builders = [{'name': 'Config Error',
                              'steps': [{'name': 'show stack',
                                         'command': 'echo %s;exit 1' % e}]}]

    def _read_config(self, conf_in):
        localDict = {
            'basedir': os.path.expanduser('.'),
            '__file__': os.path.abspath('nothinghere'),
        }
        old_sys_path = sys.path[:]
        sys.path.append(localDict)
        try:
            exec conf_in in localDict
        finally:
            sys.path[:] = old_sys_path

        builders = localDict['builders']
        return {'builders': builders}

    def getBuilders(self):
        builders = []
        for b in self.config['builders']:
            builder = copy(b)
            builder['steps'] = self._getSteps(b)
            builders.append(builder)
        return builders

    def getBuilder(self, builder_name):
        for builder in self.builders:
            if builder['name'] == builder_name:
                return builder

        msg = 'The builder named "%s" was not found. ' % builder_name
        msg += 'Maybe a bad config file?'
        raise BuilderNotFound(
            )

    def getBuildersForBranch(self, branch):
        builders = []
        for builder in self.builders:
            if branch == builder.get('branch', 'master'):
                    builders.append(builder)

        return builders

    def _getSteps(self, builder):
        steps = builder['steps']
        steps_list = []
        for step in steps:
            s = copy(step)
            s['command'] = step['command'].split()
            steps_list.append(s)
        return steps_list


class MasterConfig(MasterConfigBase):
    def __init__(self):
        MasterConfigBase.__init__(self)
        self.db.update({'toxicbuild_db_url': 'sqlite:///toxicbuild.sqlite'})

    def check_single_master(self):  # pragma: no cover
        pass

    def load_db(self, filename, config_dict):
        MasterConfigBase.load_db(self, filename, config_dict)
        if 'db' in config_dict:
            if 'toxicbuild_db_url' in config_dict['db']:
                db_url = config_dict['db']['toxicbuild_db_url']
                self.db['toxicbuild_db_url'] = db_url


class DynamicBuilderConfig(BuilderConfig):
    def __init__(self, venv_path='venv', pyversion='/usr/bin/python3.4',
                 **kwargs):
        # argh!
        from toxicbuild.process.factory import DynamicBuildFactory

        factory = DynamicBuildFactory(venv_path, pyversion)
        kwargs['factory'] = factory
        BuilderConfig.__init__(self, **kwargs)



# -*- coding: utf-8 -*-
from twisted.internet import defer
from buildbot.changes.gitpoller import GitPoller as GitPollerBase
from toxicbuild import master
from toxicbuild.config import ConfigReader
from toxicbuild.process.builder import BuilderManager

STEPS_FILE = 'toxicbuild.conf'
# :/
WRONG_CONFIG_TOKEN = 'BADCONF'


class GitPoller(GitPollerBase):
    def __init__(self, *args, **kwargs):
        GitPollerBase.__init__(self, *args, **kwargs)
        self.revList = []

    @defer.inlineCallbacks
    def poll(self):
        yield GitPollerBase.poll(self)
        for branch, rev in self.lastRev.items():
            conf = yield self._save_revconf(rev, branch)
            config = ConfigReader(conf)
            manager = BuilderManager(self.master, config)
            manager.updateBuilders()

    @defer.inlineCallbacks
    def _process_changes(self, newRev, branch):
        lastRev = self.lastRev.get(branch)
        yield self._make_revList(lastRev, newRev)
        for rev in self.revList:
            yield self._save_revconf(rev, branch)

        yield GitPollerBase._process_changes(self, newRev, branch)

    def _set_revList(self, revList):
        self._revList = revList

    def _get_revList(self):
        return self._revList

    revList = property(_get_revList, _set_revList)

    @defer.inlineCallbacks
    def _make_revList(self, lastRev, newRev):  # pragma: no cover
        # copy/paste from buildbot
        if lastRev:
            revListArgs = [r'--format=%H', '%s..%s' % (lastRev, newRev), '--']
            results = yield self._dovccmd('log', revListArgs,
                                          path=self.workdir)
            # process oldest change first
            self.revList = results.split()
            self.revList.reverse()
        yield self.revList

    @defer.inlineCallbacks
    def _save_revconf(self, revision, branch):

        try:
            conf = yield self._dovccmd('show',
                                       ['%s:%s' % (revision, STEPS_FILE)],
                                       path=self.workdir)
        except:
            conf = WRONG_CONFIG_TOKEN

        master.TOXICDB.revisionconfig.saveRevisionConfig(revision, branch,
                                                         self.repourl, conf)
        defer.returnValue(conf)



from sqlalchemy import (Table, Column, Integer, String, Text, TIMESTAMP,
                        MetaData)

meta = MetaData()

revision_config = Table(
    'revision_config', meta,
    Column('id', Integer, primary_key=True),
    Column('repourl', String(128)),
    Column('revision', String(128)),
    Column('branch', String(64)),
    Column('config', Text),
    Column('timestamp', TIMESTAMP),
)


def upgrade(migrate_engine):
    # Upgrade operations go here. Don't create your own engine; bind
    # migrate_engine to your metadata
    meta.bind = migrate_engine
    revision_config.create()


def downgrade(migrate_engine):
    # Operations to reverse the above upgrade go here.
    meta.bind = migrate_engine
    revision_config.drop()



# -*- coding: utf-8 -*-

from twisted.internet import defer
from buildbot.schedulers.base import BaseScheduler
from toxicbuild.config import ConfigReader


# yeah, dirty and ugly. but works and i don't need to write it twice!
# propably I should replace it for a proper BuilderManager.
@defer.inlineCallbacks
def addBuildsetForChanges(instance, reason='', external_idstring=None,
                          changeids=[], builderNames=None, properties=None):
    builders = yield getBuildersForChanges(instance, changeids)
    builderNames = [b['name'] for b in builders]

    rev = yield BaseScheduler.addBuildsetForChanges(
        instance, reason, external_idstring, changeids, builderNames,
        properties)

    defer.returnValue(rev)


@defer.inlineCallbacks
def getBuildersForChanges(instance, changeids):
    lastChange = yield getLastChange(instance, changeids)
    revconf = yield instance.master.toxicdb.revisionconfig.getRevisionConfig(
        lastChange['branch'], lastChange['repository'],
        lastChange['revision'])

    config = ConfigReader(revconf.config)
    builders = config.getBuildersForBranch(lastChange['branch'])
    defer.returnValue(builders)


@defer.inlineCallbacks
def getLastChange(instance, changeids):
    changesByCodebase = {}

    def get_last_change_for_codebase(codebase):
        return max(changesByCodebase[codebase],
                   key=lambda change: change["changeid"])

    for changeid in changeids:
        chdict = yield instance.master.db.changes.getChange(changeid)
        # group change by codebase
        changesByCodebase.setdefault(chdict["codebase"], []).append(chdict)

    lastchangeid = None
    for codebase in instance.codebases:
        if codebase in changesByCodebase:
            lastchangeid = get_last_change_for_codebase(codebase)['changeid']

    lastChange = yield instance.master.db.changes.getChange(lastchangeid)
    defer.returnValue(lastChange)



# -*- coding: utf-8 -*-

import os
import subprocess


def start(config):
    slave_dir = os.path.join(config['basedir'], 'slave')
    master_dir = os.path.join(config['basedir'], 'master')
    master_cmd = ['buildbot', 'start', master_dir]
    slave_cmd = ['buildslave', 'start', slave_dir]
    errors = subprocess.call(master_cmd)
    if errors:
        return 1

    errors = subprocess.call(slave_cmd)
    if errors:
        return 1

    return 0



# -*- coding: utf-8 -*-

from buildbot.process.factory import BuildFactory
from toxicbuild.process.build import DynamicBuild


class DynamicBuildFactory(BuildFactory):
    buildClass = DynamicBuild

    def __init__(self, venv_path, pyversion, steps=None):
        self.venv_path = venv_path
        self.pyversion = pyversion
        BuildFactory.__init__(self, steps)

    def newBuild(self, requests):
        b = BuildFactory.newBuild(self, requests)
        b.venv_path = self.venv_path
        b.pyversion = self.pyversion
        return b



# -*- coding: utf-8 -*-

import sys
from twisted.python import usage, reflect
from buildbot.scripts import base
from buildbot.scripts.runner import Options


class CreateToxicbuildOptions(base.BasedirMixin, base.SubcommandOptions):
    subcommandFunction = "toxicbuild.scripts.create.create"

    optFlags = [['quiet', 'q', 'Do not emit the commands being run']]

    optParameters = [
        ["toxicbuild-db", None, "sqlite:///toxicbuild.sqlite",
         "which DB to use for scheduler/status state. See below for syntax."]]

    def getSynopsis(self):
        return 'Usage: toxicbuild create <basedir>'


class StartToxicBuildOptions(base.BasedirMixin, base.SubcommandOptions):
    subcommandFunction = 'toxicbuild.scripts.start.start'

    optFlags = [['quiet', 'q', 'Do not emit the commands being run']]

    def getSynopsis(self):
        return 'Usage: toxicbuild toxicstart <basedir>'


class StopToxicBuildOptions(base.BasedirMixin, base.SubcommandOptions):
    subcommandFunction = 'toxicbuild.scripts.stop.stop'

    optFlags = [['quiet', 'q', 'Do not emit the commands being run']]

    def getSynopsis(self):
        return 'Usage: toxicbuild toxicstop <basedir>'


Options.subCommands.append(['create', None, CreateToxicbuildOptions,
                            "create easy buildbot install"])
Options.subCommands.append(['toxicstart', None, StartToxicBuildOptions,
                            'start master and slave'])
Options.subCommands.append(['toxicstop', None, StopToxicBuildOptions,
                            'stop master and slave'])


def run():  # pragma: no cover
    # copy/paste from buildbot
    config = Options()
    try:
        config.parseOptions(sys.argv[1:])
    except usage.error, e:
        print "%s:  %s" % (sys.argv[0], e)
        print
        c = getattr(config, 'subOptions', config)
        print str(c)
        sys.exit(1)

    subconfig = config.subOptions
    subcommandFunction = reflect.namedObject(subconfig.subcommandFunction)
    sys.exit(subcommandFunction(subconfig))



# -*- coding: utf-8 -*-

from twisted.python import log
from twisted.internet import defer, reactor
from buildbot.master import BuildMaster as BuildMasterBase
from toxicbuild import config
from toxicbuild.db import connector

# tragic hack to gimme access to a db connector every where, any time.
TOXICDB = None


class BuildMaster(BuildMasterBase):
    def __init__(self, *args, **kwargs):
        BuildMasterBase.__init__(self, *args, **kwargs)

        self.config = config.MasterConfig()

        self.toxicdb = connector.DBConnector(self, self.basedir)
        self.toxicdb.setServiceParent(self)
        global TOXICDB
        TOXICDB = self.toxicdb

    @defer.inlineCallbacks
    def startService(self, _reactor=reactor):
        BuildMasterBase.startService(self, _reactor=reactor)
        try:
            yield self.toxicdb.setup()
        except connector.connector.DatabaseNotReadyError:
            # (message was already logged)
            _reactor.stop()
            return

        # polling early to have some builders as soon as posible
        for poller in self.config.change_sources:
            yield poller.doPoll()

        log.msg("Toxicbuild is running")



# -*- coding: utf-8 -*-

from buildbot.db import base


class RevisionConfigConnectorComponent(base.DBConnectorComponent):

    def saveRevisionConfig(self, revision, branch, repourl, config):
        def thd(conn):
            return self._saveRevisionConfig(
                conn, revision, branch, repourl, config)

        return self.db.pool.do(thd)

    def _saveRevisionConfig(self, conn, revision, branch, repourl, config):
        # yeah, this method is silly, but there's a reason.
        # somethimes I really need to block the whole thing
        r = conn.execute(self.db.model.revisionconfig.insert(),
                         dict(revision=revision, branch=branch,
                              repourl=repourl, config=config))

        return r.inserted_primary_key[0]

    def getRevisionConfig(self, branch, repourl=None, revision=None):
        def thd(conn):
            return self._getRevisionConfig(conn, branch, repourl, revision)

        return self.db.pool.do(thd)

    def _getRevisionConfig(self, conn, branch, repourl=None, revision=None):
        # the silly thing again
        tbl = self.db.model.revisionconfig
        query = tbl.select(whereclause=(tbl.c.branch == branch))
        if repourl:
            query = query.where(tbl.c.repourl == repourl)

        if revision:
            query = query.where(tbl.c.revision == revision)
        else:
            query = query.order_by('-id')

        r = conn.execute(query)
        return r.fetchone()



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

VERSION = '0.2'
DESCRIPTION = """
Hackish pieces of software to an easy buildbot config
"""
LONG_DESCRIPTION = DESCRIPTION

setup(name='toxicbuild',
      version=VERSION,
      author='Juca Crispim',
      author_email='jucacrispim@gmail.com',
      url='http://toxicbuild.readthedocs.org/en/latest/',
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      packages=find_packages(exclude=['tests', 'tests.*']),
      include_package_data=True,
      install_requires=['buildbot>=0.8.9', 'buildbot-slave>=0.8.9'],
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: No Input/Output (Daemon)',
          'Environment :: Web Environment',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU General Public License (GPL)',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Topic :: Software Development :: Build Tools',
          'Topic :: Software Development :: Testing',
      ],
      scripts=['script/toxicbuild'],
      test_suite='tests',
      provides=['toxicbuild'],)



# -*- coding: utf-8 -*-

import copy
from twisted.python import log
from buildbot.process.builder import Builder
from buildbot.schedulers.forcesched import ForceScheduler
from toxicbuild.config import DynamicBuilderConfig


class BuilderManager:
    def __init__(self, master, config, category=None):
        self.master = master
        self.config = config
        self.category = category

    def updateBuilders(self):
        """
        Creates new builders and remove old ones based on self.config
        """

        removed = self.removeOldBuilders()
        builders = self.createNewBuilders()
        self.updateSlaves()
        return builders, removed

    def createNewBuilders(self):

        builders = {}
        for bdict in self.config.builders:
            if bdict['name'] in self.master.botmaster.builderNames:
                continue

            log.msg('builder: creating builder %s' % bdict['name'])

            force = bdict.get('forceScheduler', True)
            builder = self.createBuilderFromDict(bdict)
            builders[bdict['name']] = builder
            if force:
                self.addToForceScheduler(builder)

        builderNames = [b['name'] for b in self.config.builders]
        self.master.botmaster.builders.update(builders)
        self.master.botmaster.builderNames = self.master.botmaster.\
            builders.keys()

        return builderNames

    def updateSlaves(self):
        for s in self.master.config.slaves:
            s.updateSlave()

    def removeOldBuilders(self):
        builderNames = copy.copy(self.master.botmaster.builderNames)
        new_builders = [b['name'] for b in self.config.builders]
        removed = []
        for b in builderNames:
            if b in new_builders:
                continue

            log.msg('builder: removing builder %s' % b)
            removed.append(b)
            self.master.status.builderRemoved(b)
            self.master.botmaster.builderNames.pop(
                self.master.botmaster.builderNames.index(b))
            builder = self.master.botmaster.builders[b]
            builder.disownServiceParent()
            del self.master.botmaster.builders[b]

        return removed

    def createBuilderFromDict(self, bdict):
        builder = Builder(bdict['name'])
        keys2delete = ['steps', 'branch', 'forceScheduler', 'candies']
        for key in keys2delete:
            try:
                del bdict[key]
            except KeyError:
                pass

        if 'slavenames' not in bdict.keys():
            bdict['slavenames'] = [s.slavename for s in
                                   self.master.config.slaves]

        bconf = DynamicBuilderConfig(**bdict)
        builder.config = bconf
        builder.master = self.master
        builder.botmaster = self.master.botmaster
        builder.builder_status = self.master.status.builderAdded(
            bconf.name,
            bconf.builddir,
            bconf.category,
            bconf.description)

        builder.setServiceParent(builder.botmaster)
        return builder

    def addToForceScheduler(self, builder):
        """
        Sets the builder in all force schedulers configured
        in master
        """

        for sched in self.master.config.schedulers.values():
            if isinstance(sched, ForceScheduler):
                sched.builderNames.append(builder.name)



# -*- coding: utf-8 -*-

import os
import jinja2
from pkg_resources import resource_filename
from twisted.internet import defer
from twisted.python import util
from buildbot.util import in_reactor
from buildbot.scripts.create_master import (makeBasedir,
                                            makePublicHtml,
                                            makeTemplatesDir, createDB)
from buildslave.scripts.create_slave import createSlave
from migrate.versioning.api import version_control, test, upgrade


@in_reactor
@defer.inlineCallbacks
def create(config):
    create_basedir(config)
    master_config = _get_master_config(config)
    yield createMaster(master_config)
    yield createToxicbuildDB(master_config)

    slave_config = _get_slave_config(config)
    createSlave(slave_config)
    if not config['quiet']:
        print('master configured at %s' % master_config['basedir'])
        print('slave configured at %s' % slave_config['basedir'])
    defer.returnValue(0)


@defer.inlineCallbacks
def createToxicbuildDB(config):
    repository = resource_filename('toxicbuild', 'migrations')
    url = config['toxicbuild-db']
    if not config['quiet']:
        verbose_url = url.replace(config['basedir'], '')
        print('creating database for toxicbuild (%s)' % verbose_url)

    version_control(url, repository)
    test(url, repository)
    yield upgrade(url, repository)


def create_basedir(config):
    pythonpath = os.path.join(config['basedir'], 'pythonpath')
    if not config['quiet']:
        print('mkdir %s' % config['basedir'])
        print('mkdir %s' % pythonpath)

    os.mkdir(config['basedir'])
    os.mkdir(pythonpath)


@defer.inlineCallbacks
def createMaster(config):  # pragma: no cover
    # copy/paste from buildbot.
    makeBasedir(config)
    makeTAC(config)
    makeSampleConfig(config)
    makePublicHtml(config)
    makeTemplatesDir(config)
    yield createDB(config)


def makeSampleConfig(config):  # pragma: no cover
    # copy/paste from buildbot
    source = util.sibpath(__file__, "sample.cfg")
    target = os.path.join(config['basedir'], "master.cfg")
    if not config['quiet']:
        print "creating %s" % target
    with open(source, "rt") as f:
        config_sample = f.read()
    if config['db']:
        config_sample = config_sample.replace('sqlite:///state.sqlite',
                                              config['db'])
    with open(target, "wt") as f:
        f.write(config_sample)
    os.chmod(target, 0600)


def makeTAC(config):  # pragma: no cover
    # copy/paste from buildbot
    # render buildbot_tac.tmpl using the config
    loader = jinja2.FileSystemLoader(os.path.dirname(__file__))
    env = jinja2.Environment(loader=loader, undefined=jinja2.StrictUndefined)
    env.filters['repr'] = repr
    tpl = env.get_template('buildbot_tac.tmpl')
    cxt = dict((k.replace('-', '_'), v) for k, v in config.iteritems())
    contents = tpl.render(cxt)

    tacfile = os.path.join(config['basedir'], "buildbot.tac")
    if os.path.exists(tacfile):
        with open(tacfile, "rt") as f:
            oldcontents = f.read()
        if oldcontents == contents:
            if not config['quiet']:
                print "buildbot.tac already exists and is correct"
            return
        if not config['quiet']:
            print "not touching existing buildbot.tac"
            print "creating buildbot.tac.new instead"
        tacfile += ".new"
    with open(tacfile, "wt") as f:
        f.write(contents)


def _get_master_config(config):
    master_config = {}
    master_config['basedir'] = os.path.join(config['basedir'], 'master')
    master_config['quiet'] = config['quiet']
    master_config['force'] = False
    master_config['relocatable'] = True
    master_config['no-logrotate'] = False
    master_config['config'] = 'master.cfg'
    master_config['log-size'] = "10000000"
    master_config['log-count'] = "10"
    master_config['db'] = "sqlite:///state.sqlite"
    master_config['toxicbuild-db'] = _get_toxicbuild_db_url(
        config['toxicbuild-db'], master_config['basedir'])

    return master_config


def _get_slave_config(config):
    slave_config = {}
    slave_config['basedir'] = os.path.join(config['basedir'], 'slave')
    slave_config['quiet'] = False
    slave_config['force'] = False
    slave_config['relocatable'] = True
    slave_config['no-logrotate'] = False
    slave_config['log-size'] = 10000000
    slave_config['log-count'] = 10
    slave_config['umask'] = "None"
    slave_config['usepty'] = 0
    slave_config['keepalive'] = 600
    slave_config['maxdelay'] = 300
    slave_config['allow-shutdown'] = None
    slave_config['name'] = 'easyslave'
    slave_config['passwd'] = 'dummypass'
    slave_config['host'] = 'localhost'
    slave_config['port'] = 9989

    return slave_config


def _get_toxicbuild_db_url(toxicbuild_db, basedir):
    origurl = url = toxicbuild_db

    if origurl.startswith('sqlite'):
        path = origurl.split('sqlite:///')[1]
        if not path.startswith('/'):
            url = os.path.join(basedir, path)
            url = 'sqlite:///' + url

    return url



# -*- coding: utf-8 -*-

import sqlalchemy as sa
from buildbot.db import base


class Model(base.DBConnectorComponent):
    metadata = sa.MetaData()

    revisionconfig = sa.Table(
        'revision_config', metadata,
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('repourl', sa.String(128)),
        sa.Column('revision', sa.String(128)),
        sa.Column('branch', sa.String(64)),
        sa.Column('config', sa.Text),
        sa.Column('timestamp', sa.TIMESTAMP),

        sa.UniqueConstraint('repourl', 'revision')
    )
